/* ---------------------------------------------------------------------
 *       killdef filename MACRO_NAME [ [MACRO_NAME2] ...]
 * Imitates the C Preprocessor partly:
 * Process <filename> assuming that <MACRO_NAME> is _not_ defined.
 * Input is <filename>, Output goes to STDOUT.
 * compile with
 *      cc -s -o killdef killdef.c
 * ---------------------------------------------------------------------
 * This file is a part of tcharge, version 1.3, 11MAY98
 * Copyright (C) 1997,98 Volker B"orchers (boercher@physik.uni-bremen.de)
 * See file `tcharge.c' for legal affairs.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define USE "Usage:\n\tkilldef filename MACRO_NAME [ [MACRO_NAME2] ...]\
\nImitate the C Preprocessor partly:\
\nProcess <filename> assuming that <MACRO_NAME> is _not_ defined.\
\nInput is <filename>, Output goes to STDOUT.\n"

FILE *my_fopen(const char name[], const char mode[]);

int main(int argc, char **argv)
{
  FILE *inf, *outf;
  int level = 0;   /* level of the brackets */
  int searchlevel = -1;		/* if #ifdef is concerned */
  int invsearchlevel = -1;  	/* if #ifndef is concerned */
  int offset, count = 0;
  int i;
  short doprint = 1;
  short inverse = 0;  
  char line[512];
  
  if(argc < 3) {
    fprintf(stderr, USE);
    exit(1);
  }
  
  if((inf = my_fopen(argv[1], "rt")) == NULL) {
    fprintf(stderr, "Cannot open `%s\'", argv[1]);
    exit(1);
  }
  outf= stdout;
  
  
  while(!feof(inf)) {
    fgets(line, 511, inf);	/* read in the next line */
    if(line[0] == '#') {	/* preprocessor direction */
      for(offset = 1; isspace(line[offset]); offset++)
	;  /* there might be white space between # and .. */
      if(strncmp(&line[offset], "ifdef", 5) == 0) {		/* #ifdef */
	level++;
	offset += 5;
	if(doprint) {
	  for( ; isspace(line[offset]); offset++)
	    ;	/* and there might be white space before the macro */
	  for(i=2; i<argc; i++) {
	    if(strncmp(&line[offset], argv[i], strlen(argv[i])) == 0) {
	      /* compare to the list of macros */
	      if(isspace( line[ offset + strlen(argv[i]) ] )) {
		/* to distinct THIS and NOT THIS */
		count++;
		inverse = 0;
		doprint = 0;		  /* turn it of */
		searchlevel = level;
		break;
	      }
	    }
	  }
	}
      }
      else if(strncmp(&line[offset], "ifndef", 6) == 0) {	/* #ifndef */
	level++;
	offset += 6;
	if(doprint) {
	  for( ; isspace(line[offset]); offset++)
	    ;	/* and there might be white space before the macro */
	  for(i=2; i<argc; i++) {
	    if(strncmp(&line[offset], argv[i], strlen(argv[i])) == 0) {
	      /* compare to the list of macros */
	      if(isspace( line[ offset + strlen(argv[i]) ] )) {
		/* to distinct THIS and NOT THIS */
		count++;
		invsearchlevel = level;
		strcpy(line, "\0");
		break;
	      }
	    }
	  }
	}
      }
      else if(strncmp(&line[offset], "if", 2) == 0) {		/* #if EXPRESSION */
	/* Don't worry about this, we don't want to build a real CPP! */
	level++;
      }
      else if(strncmp(&line[offset], "else", 4) == 0) {	/* #else */
	if(level == searchlevel) {
	  doprint = 1; 
	  strcpy(line, "\0");	/* don't print the leading #else */
	}
	else if(level == invsearchlevel) {
	  doprint = 0; 
	  strcpy(line, "\0");	/* don't print the leading #else */
	}
      }
      else if(strncmp(&line[offset], "endif", 5) == 0) {	/* #endif */
	level--;
	if(level == searchlevel-1) {
	  doprint = 1;		/* turn it on */
	  searchlevel = -1;
	  strcpy(line, "\0");
	}
	else if(level == invsearchlevel-1) {
	  doprint = 1;		/* turn it on */
	  invsearchlevel = -1;
	  strcpy(line, "\0");
	}
      }
    }
    if(doprint)
      fputs(line, outf);
  }
  if(level != 0) {
    fprintf(stderr, "Parsing error: %d open brackets\n", level);
    /* under some circumstances this 
     * don't works near the EOF!!! */
    return 1;
  }
  else
    fprintf(stderr, "%d blocks deleted\n", count);
  return 0; 	/* return successfully */
}







FILE *my_fopen(const char name[], const char mode[])
{
  FILE *fpointer;	
  
  if((fpointer = fopen(name, mode)) == NULL) {
    fprintf(stderr, "* Cannot open file %s.\n", name);
    exit(0);
  }
  
  return(fpointer);
}

