/*
 * tcharge, version 1.3 11MAY98
 * tcharge.c: Source for 'tcharge' - modem charge watch for X
 *
 * Copyright (C) 1997,98 Volker B"orchers (boercher@physik.uni-bremen.de)
 * This file is part of tcharge
 *
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or (at 
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
 * for more details.
 * 
 * Feel free to make changes to tcharge. The author invites you
 * explicitely to add support for other telephone companies.
 * If you do that please send a report to the author.
 *
 * =====================================================================
 * Essentially all X programming of tcharge is taken from rclock,
 * version 1.81. RCLOCK: Copyright 1992, 1993 Robert Nation
 * (nation@rocket.sanders.lockheed.com)
 * =====================================================================
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string.h>
#include <signal.h>
#include <syslog.h>
#ifdef ISDN
#include <fcntl.h>
#endif
/*==================== NATIONAL SPECIFIC (BEGIN) ======================*/
/*
 * all charge informations and nat. spec. defined in charge.h, a symlink to
 * the appropriate header file in ./National.
 */
#include "charge.h"
/*
 * Setting the defaults for (text) display.
 */
#ifdef BILLING_PER_UNIT
#  define OPT_UNITS_DEFAULT 1	/* display spent units */
#  define OPT_CHARGE_DEFAULT 0	/* do not print the current charge */
#else /* BILLING_PER_SECOND: */
#  define OPT_UNITS_DEFAULT 0	/* does not make sense if BILLING_PER_SECOND */
#  define OPT_CHARGE_DEFAULT 1	/* display of the current charge */
#endif /* BILLING_PER_UNIT || BILLING_PER_SECOND */
/* 
 * the help string for the zone depends is national specific too,
 * (and should be set in the chargefile). In case it's undefined:
 */
#ifndef ZONE_HELP_STRING
#define ZONE_HELP_STRING "-zone <n>         n = 0) Local (default), for other\n\
                      see definitions for your telephone company"
#endif
/*===================== NATIONAL SPECIFIC (END) =======================*/

char *tcharge_pid_file = "/var/run/tcharge.pid";
#define TCHARGE_VERSION         "1.3"
#define CLASS	        "Clock"
#define MESS_CLASS	"Warning"

#define MW_EVENTS   (ExposureMask | StructureNotifyMask | ButtonPressMask)
#define MESS_EVENTS (ExposureMask | StructureNotifyMask | ButtonPressMask)

/* screen info */
Display		*display;
int             screen;		   /* the X screen number */
int             MyDisplayWidth, MyDisplayHeight;
int             x_fd;              /* file descriptor of X server connection */

/* windows and their sizes */
Window		main_win;               /* parent window */
int             main_height=100, main_width = 95;

#ifdef REMINDERS
Window		mess_win;                /* message window */
int             mess_w,mess_h;
#endif

/* Two variables for reset the focus to the parent window of tcharge: */
Window focus;
int revert_to;

/* graphics info accessible by commandline options */
char *display_name = NULL;
char *bg_color = "white";
char *fg_color = "black";
char *geom_string = "95x100";

/* assorted graphics info */
GC 		gc;	                 /* GC for normal video */
GC              rgc;                     /* GC for reverse video */

GC              mgc;                     /* GC for text */
XFontStruct     *font;
int             font_height;
char            *font_string = "fixed";

XColor          background_color;
XColor          foreground_color;
unsigned long	current_bg_pixel;	/* the current background color */
#define BW 1                            /* Window border width */

char            *app_name = "tcharge";      /* the resource name */
#ifdef REMINDERS
char            *mess_name = MESS_CLASS;
int             Mapped = 0;             /* is message win mapped? */
char            message[256]="";
#endif

int             clockupdate = 15;
int		clockupdate_save = 15; /* Needed in "demon mode" */

static XSizeHints sizehints = {
	PMinSize | PResizeInc | PBaseSize | PWinGravity,
	0, 0, 95, 100,	/* x, y, width and height */
	1, 22,		/* Min width and height */
	0, 0,		/* Max width and height */
	1, 1,		/* Width and height increments */
	{1, 1},		/* x,y increments */
	{1, 1},		/* Aspect ratio */
	0, 0 ,		/* base size */
	NorthWestGravity
};

/* save from computing sine(x), use pre-computed values
 * They are *100, to avoid using floats */
short sine[]={0,105,208,309,407,500,588,669,743,809,866,914,951,
	      978,995,999,995,978,951,914,866,809,743,669,588,500,
	      407,309,208,105,0,-105,-208,-309,-407,-500,-588,-669,
	      -743,-809,-866,-914,-951,-978,-995,-999,-994,-978,-951,
	      -914,-866,-809,-743,-669,-588,-500,-407,-309,-208,-105 };

/* subroutine declarations */
void XMode(int, char **);
void get_options(int, char **);
void print_charge_structure(void);
void write_pid_file(void);
void extractResources(char *, char *, char *);
void createWindows(int,char **);
void getEvent(void);
void getEvent_noX(void);
void DrawWindow(Window, int, int, int, int);
int  CalcUpTime(void);
int  connected(void);
void setstarttime (time_t start_t);
#ifdef REMINDERS
void SetOffAlarm(void);
#endif
void print_summary(void);
static void signal_handler __P ((int));
void error(char *);

/* OPTIONS */
int opt_noclock=0;	/* 1: display no clock (X-text only) */
int opt_noX=0;		/* 1: no X-window at all (logfile only) */
int opt_zone_number=0;	/* phone charge zones */
int opt_before=0;	/* really got connected 'opt_before' sec. earlier */
int opt_after=0;	/* if we would hang up now, conn. terminates 'opt_after' sec. later */
int opt_units=OPT_UNITS_DEFAULT;	/* display spent units? */
int opt_charge=OPT_CHARGE_DEFAULT;	/* display charge? */
int opt_warn_seconds=0;	/* seconds to warn before next unit begins */
int opt_bell=0;		/* choice for sort of alarm (beep or blink) */
char user_string[2*L_cuserid+9] = ""; /* The (e)userid - string to be printed (`-user') */
int opt_syslog=0;	/* if print_summary() should use syslog */
time_t opt_start=0;	/* For test purpose: Start time (seconds since 1.1. 1970) */
time_t opt_end=0;	/* For test purpose: End time (seconds since 1.1. 1970) */

/* phone line statistics */
struct tm *start, start_save;	/* start time, start is a pointer on start_save */
int total_dur;		/* total duration of the call */
float charge;		/* the charge in [CURRENCY_NAME] */
float old_charge=0.0;	/* the last charge in [CURRENCY_NAME] */
float ten_to_the_decimals; /* for charge upward rounding in CalcUpTime() */
int i_units;		/* number of units */
int old_units=-1;	/* last value of unit number */
int today, tomorrow;	/* flags indicating the day is a holiday */
char time_str[2][32]={"",""}; /* main window text (maybe 2 lines) */
char charge_str[32];	/* the printf string for the text-display of charge */
int text_lines;		/* normally there is only 1 text line */
char *summaryfile;	/* The name of the file to print statistics to */
char *watch_file;	/* File to watch (start/stop counting if it's created/deleted) */
int halted = 0;		/* signals if tcharge is stopped */

#define BLINKTIME 500000 /* time in reversed video when blinking in 1/1000000 seconds */
#define VOLUME -10

#ifdef REMINDERS
#define WARN_CHARGE 1
#define WARN_MIN 2
#define WARN_UNITS 3
float opt_warntime;	/* (float) charge || (int) minutes or units */
float warntime;		/* = opt_warntime * 1,2,3,... */
int opt_warn = 0;	/* values: 0 || WARN_XXX */
#endif /* REMINDERS */


/****************************************************************************
 *
 * tcharge -- main()
 *
 ****************************************************************************/
void main(int argc, char **argv)
{
# ifndef BILLING_PER_UNIT
  int i;
  for(i=0, ten_to_the_decimals=1.0; i<DECIMALS; i++) /* if BILLING_PER_SECOND */
    ten_to_the_decimals *= 10.0;
# endif
# ifdef CURRENCY_AFTER_CHARGE
  sprintf(charge_str, "%%.%df %s", DECIMALS, CURRENCY_NAME);
# else
  sprintf(charge_str, "%s %%.%df", CURRENCY_NAME, DECIMALS);
# endif
  /* Catch signals SIGHUP, SIGINT, SIGQUIT, SIGPIPE or SIGTERM
   * and execute signal_handler() before doing the default actions.
   * Note: SIGKILL and SIGSTOP cannot be caught!
   */
  signal(SIGHUP, signal_handler);
  signal(SIGINT, signal_handler);
  signal(SIGQUIT, signal_handler);
  signal(SIGTERM, signal_handler);
  signal(SIGPIPE, signal_handler);
    
  /* parse the command line */
  get_options(argc, argv);
  
  /* set zone-specific variables (see charge.h) */
  init_zone[opt_zone_number]();

  /* Option -end forces tcharge to terminate after one calculation */
  if(opt_end) {
    setstarttime (time(0));
    CalcUpTime();
    /* hope the signal handler is properly installed! */
    kill (getpid (), SIGTERM);
  }

  /* create a pid file (tcharge_pid_file) if possible */
  write_pid_file();

  if(! opt_noX)
    XMode(argc, argv);
  else
    getEvent_noX();
  
  if(tcharge_pid_file != NULL)
    unlink(tcharge_pid_file);
}

/*****************************************************************************
 *
 *  Initialize all X-stuff and loop forever
 *
 *****************************************************************************/
void XMode(int argc, char **argv)
{
  XGCValues gcv;
  
  if((display_name = getenv ("DISPLAY")) == NULL)
    display_name = ":0";
  
  text_lines = opt_units + opt_charge;
  if(text_lines < 1 || text_lines > 2)
    error("Wrong number of textlines: must be in [1,2]");
    
  /* Initialize time */
  if(watch_file == NULL)
    setstarttime (time(0));
  /* open display */
  if (!(display = XOpenDisplay(display_name)))
    error("Can't open display");
  /* get current focus settings: */
  XGetInputFocus(display, &focus, &revert_to);
  
  /* get display info */
  x_fd = XConnectionNumber(display);
  screen = DefaultScreen(display);
  MyDisplayWidth = DisplayWidth(display, screen);
  MyDisplayHeight = DisplayHeight(display, screen);
  
  extractResources(geom_string, fg_color, bg_color);
  createWindows(argc,argv);
  /* load the font for messages */
  if ((font = XLoadQueryFont(display,font_string)) == NULL) {
    fprintf(stderr,"can't access font %s\n",font_string);
    error("Exiting");
  }
  font_height = font->ascent + font->descent;
  /* redeclare the minimum window height */
  sizehints.min_height = font_height + 15;
  /*  Create the graphics contexts. */
  gcv.foreground = foreground_color.pixel;
  gcv.background = background_color.pixel;
  gcv.font = font->fid;
  gc = XCreateGC(display,main_win,GCFont|GCForeground|GCBackground,&gcv);
  current_bg_pixel = background_color.pixel; /* for use in DrawWindow() */
#   ifdef REMINDERS
  mgc = gc;
#   endif
  gcv.foreground = background_color.pixel;
  gcv.background = foreground_color.pixel;
  rgc = XCreateGC(display,main_win,GCFont|GCForeground|GCBackground,&gcv);
  
  getEvent();
}


/*****************************************************************************
 *
 *  Parse the command line
 *
 *****************************************************************************/
void get_options(int argc, char **argv)
{
  int i;

  for(i=1;i<argc;i+=2) {
    if(*(++argv[i]) == '-') argv[i]++;  /* accept GNU-style long options (--) */
    if(strcmp(argv[i],"display")==0)
      display_name = argv[i+1];
    else if(strcmp(argv[i],"geometry")==0)
      geom_string = argv[i+1];
    else if(strcmp(argv[i],"fg")==0)
      fg_color = argv[i+1];
    else if(strcmp(argv[i],"bg")==0)
      bg_color = argv[i+1];
    else if(strcmp(argv[i],"font")==0)
      font_string = argv[i+1];
    else if(strcmp(argv[i],"noclock")==0) {
      opt_noclock = 1;
      i--;  /* since it is a switch (without arg) */
    }
    else if(strcmp(argv[i],"noX")==0) {
      opt_noX = 1;
      i--;  /* since it is a switch (without arg) */
    }
    else if(strcmp(argv[i],"holiday")==0) {
      today = 1;
      i--;  /* a switch */
    }
    else if(strcmp(argv[i],"units-only")==0) { /* a switch */
      opt_units = 1;
      opt_charge = 0;
      i--;
#     ifndef BILLING_PER_UNIT
      error("'-units_only' makes no sense in case of billing per second!");
#     endif
    }
    else if(strcmp(argv[i],"charge-only")==0) { /* a switch */
      opt_charge = 1;
      opt_units = 0;
      i--;
    }
    else if(strcmp(argv[i],"units-and-charge")==0) { /* a switch */
      opt_charge = 1;
      opt_units = 1;
      i--;
#     ifndef BILLING_PER_UNIT
      error("'-units-and-charge' makes no sense in case of billing per second!");
#     endif
    }
    else if(strcmp(argv[i],"warn-seconds")==0) {
      if (sscanf(argv[i+1],"%d",&opt_warn_seconds)!=1) {
	fprintf(stderr, "Wrong '-warn-seconds' value: %s\n",argv[i+1]);
	error("exiting");
      }
#     ifndef BILLING_PER_UNIT
      error("'-warn-seconds' makes no sense in case of billing per second!\n");
#     endif
    }
    else if(strcmp(argv[i],"bell")==0) { /* a switch */
      opt_bell = 1;
      i--;
    }
#   ifdef REMINDERS
    else if(strcmp(argv[i],"warn")==0) {
      char tmps[16];
      if(sscanf(argv[i+1], "%f%s", &opt_warntime, tmps) != 2) {
	fprintf(stderr,
		"tcharge: Wrong argument to '-warn': %s\nNo Warnings!\n",
		argv[i+1]);
	continue;
      }
      warntime = opt_warntime;
      if(strncmp(tmps, CURRENCY_NAME, 2) == 0)
	opt_warn = WARN_CHARGE;	/* currency "warntimes" are float */
      else {				/* other "warntimes" are int */
	if(strncmp(tmps, "min", 3) == 0)	/* minute(s) */
	  opt_warn = WARN_MIN;
	else if(strncmp(tmps, "unit", 4) == 0) /* unit(s) */
	  opt_warn = WARN_UNITS;
	else {
	  fprintf(stderr,
		  "tcharge: Wrong argument to '-warn': %s\nNo Warnings!\n",
		  argv[i+1]);
	  continue;
	}
	if((opt_warntime = (int) opt_warntime) < 1) {
	  fprintf(stderr,
		  "tcharge: -warn'time' %d %s < 1: \nNo Warnings!\n",
		  (int)opt_warntime, tmps);
	  opt_warn = 0;
	}
      }
    }
#   endif /* REMINDERS */
    else if (strcmp(argv[i],"file")==0) {
      summaryfile = argv[i+1];
    }
    else if (strcmp(argv[i],"syslog")==0) {
      opt_syslog = 1;
      i--;  /* since it is a switch (without arg) */
    }
    else if (strncmp(argv[i],"user",4)==0) {
      /* user name explicitely given? ("-user=<username>") */
      if(index(argv[i], '=') != NULL) {
	sprintf(user_string, "(user %s) ", (argv[i]+5));	
      }
      else {
	/* create a string containing the (eff.) user name to print */
	if(geteuid() != getuid()) /* if tcharge is suid print both */
	  sprintf(user_string, "(user %s/%s) ", getlogin(), cuserid(NULL));
	else
	  sprintf(user_string, "(user %s) ", cuserid(NULL));
      }
      i--;  /* since it _behaves_ like a switch (without arg) */
    }
    else if (strcmp(argv[i],"watch-file")==0) {
      watch_file = argv[i+1];
      halted = 1; /* in "demon mode" start counting only when 'watch_file' present */
#     ifdef ISDN
      error("Don't use \"-watch-file\" when compiled with -DISDN!");
#     endif /* ISDN */
    }
    else if (strcmp(argv[i],"isdn")==0) {
      watch_file = "dummy";
      halted = 1;
      /* clockupdate = 1; --> will be done after parsing the options */
#     ifndef ISDN
      error("Option \"-isdn\" not available - recompile with -DISDN!");
#     endif /* ISDN */
      i--;
    }
    else if (strcmp(argv[i],"zone")==0) {
      if (sscanf(argv[i+1],"%d",&opt_zone_number)!=1 ||
	  opt_zone_number < 0 || opt_zone_number > NUM_ZONES-1) {
	fprintf(stderr, "Wrong zone number: %s\n",argv[i+1]);
	error("exiting");
      }
    }
    else if (strcmp(argv[i],"before")==0) {
      if (sscanf(argv[i+1],"%d",&opt_before)!=1) {
	fprintf(stderr, "Wrong '-before' value: %s\n",argv[i+1]);
	error("exiting");
      }
    }
    else if (strcmp(argv[i],"after")==0) {
      if (sscanf(argv[i+1],"%d",&opt_after)!=1) {
	fprintf(stderr, "Wrong '-after' value: %s\n",argv[i+1]);
	error("exiting");
      }
    }
    else if (strcmp(argv[i],"update")==0) {
      if (sscanf(argv[i+1],"%d",&clockupdate)!=1 ||
	  clockupdate<1) {
	fprintf(stderr, "Wrong update value: %s\n",argv[i+1]);
	clockupdate=15;
      } else if (clockupdate>60) {
	fprintf(stderr, "Update value > 60sec, setting to 60sec\n");
	clockupdate=15;
      }
      clockupdate_save = clockupdate;
    }
    else if (strcmp(argv[i],"start")==0) {
      if (sscanf(argv[i+1],"%ld",&opt_start)!=1) {
	fprintf(stderr, "Wrong start time, cannot convert `%s' into long.\n",argv[i+1]);
	error("exiting");
      }
    }
    else if (strcmp(argv[i],"end")==0) {
      if (sscanf(argv[i+1],"%ld",&opt_end)!=1) {
	fprintf(stderr, "Wrong end time, cannot convert `%s' into long.\n",argv[i+1]);
	error("exiting");
      }
    }
    else if (strcmp(argv[i],"pidfile")==0) {
      tcharge_pid_file = argv[i+1];
      if (strcmp(tcharge_pid_file, "false")==0) /* unset? */
	tcharge_pid_file = NULL;
    }
    else if (strcmp(argv[i],"help")==0) {
      fprintf(stderr, "\nUsage of tcharge (version %s):\n\
-help                 print this help message and exit\n\
-version              print version informations and exit\n\
-help-charge          print the charge structure and exit\n\
-display <name>       specify the display (server)\n\
-geometry <spec>      the initial window geometry\n\
-bg <color>           background color\n\
-fg <color>           foreground color\n\
-font <font_name>     Specifies a font other than fixed to use for both,\n\
                      main and message window.
-update <n>           Update the clock face after n seconds 1<=n<=60.\n\
%s\n\
-holiday              today is a holiday: use weekend charging\n\
-before <sec>         Really got connected 'sec' seconds earlier\n\
-after <sec>          If we'ld hang up now, connection terminates 'sec' later\n\
-noclock              display no clock (text only)\n\
-noX                  No X-window at all (use option '-file')\n",
	      TCHARGE_VERSION, ZONE_HELP_STRING);
#     ifdef BILLING_PER_UNIT
      fprintf(stderr, "-units-only           print spent units (not the charge)\n\
-charge-only          print charge (not the units)\n\
-units-and-charge     print both, spent units and charge\n\
-warn-seconds <sec>   warn 'sec' seconds before next unit begins\n\
-bell                 sound a bell instead of flashing the window\n");
#     endif /* BILLING_PER_UNIT || BILLING_PER_SECOND */

#     ifdef REMINDERS
      fprintf(stderr,"-warn <Nunits|Nmin>   Warn every 'N' units or minutes spend\n\
-warn <x%s>           Warn if the charge of 'x' %s is reached\n",
              CURRENCY_NAME, CURRENCY_NAME);
#     endif /* REMINDERS */
      fprintf(stderr,"-file <filename>      \
Write phone call statistics to 'filename' (append)
-syslog               Write statistics via syslog to system logging files
-user[=<username>]    Print also the (effective) user name
-start <sec>          For test purpose: Start time (seconds since 1.1. 1970)
-end   <sec>          For test purpose:  End  time (seconds since 1.1. 1970)
-pidfile <pidfile>    Change location and/or name of the tcharge pidfile
                      (<pidfile> is \"false\", no file will be written).\n");
#     ifndef ISDN
      fprintf(stderr,"-watch-file <file>    \
\"Demon mode\": start/stop counting when 'file'\n\
                      is created/deleted\n");
#     else /* #ifdef ISDN */
      fprintf(stderr,"-isdn                 \"Demon mode\": Look out for ISDN-connections\n");
#     endif /* ISDN */
      exit(0);
    }
    else if (strcmp(argv[i],"version")==0) {
      fprintf(stderr, "Tcharge version %s by Volker B\"orchers,\
              \n Module \"%s\", charge file dated %s\n",
	      TCHARGE_VERSION, COUNTRY, CHARGE_DATE);
      exit(0);
    }
    else if (strcmp(argv[i],"help-charge")==0) {
      print_charge_structure();
      exit(0);
    }
    else {
      fprintf(stderr, "Unknown option '-%s'\n try 'tcharge -help'\n", argv[i]);
      error("exiting");
    }
  }
  /* ----- parsing command line options done! ----- */
# ifdef ISDN
  if(watch_file == NULL)
    clockupdate = 1; /* Since I don't know any other procedure for getting the
		      * ISDN-connect-time than to check /dev/isdninfo (`in connected()')
		      * every 'clockupdate' seconds, we have to ensure that this value
		      * is not to large  */
# endif
  /* ----- print some logmessages when compiled for DEBUG: ----- */
# ifdef DEBUG
#  ifndef PRINT_CHARGE_STRUCTURE /* need extended help when debugging! */
#   define PRINT_CHARGE_STRUCTURE
#  endif
  printf("This is tcharge, (version %s) compiled for %s with '-DDEBUG'\n",
	 TCHARGE_VERSION, COUNTRY);
#  ifdef ISDN
  printf("tcharge compiled with compilerflag '-DISDN'\n");
#  endif
#  ifdef REMINDERS
  printf("tcharge compiled with compilerflag '-DREMINDERS'\n");
#  endif
#  ifdef PRINT_CHARGE_STRUCTURE
  printf("tcharge compiled with compilerflag '-DPRINT_CHARGE_STRUCTURE'\n");
#  endif
  if(watch_file == NULL)
    printf("(\"Normal mode\": Start computing charge now, stop on exit of tcharge.)\n");
  else
    printf("(\"Demon mode\": Waiting for '%s' beeing created.)\n", watch_file);
# endif /* DEBUG */

  if(opt_noX) {
    if(summaryfile == NULL)
      error("Option '-noX' makes no sense without option '-file <logfile>'\n");
  }
  return;
}

/**************************************************************************
 *
 *  this function prints the charge structure as given
 *  in the header file 'charge.h'
 *
 **************************************************************************/
void print_charge_structure(void)
{
#ifdef PRINT_CHARGE_STRUCTURE
  int z, d, b; /* zone, day, band ('timezone') */
  char *info_string[DAYS_A_WEEK];
  char *unit_length;	/* header string for unit_length or charge/minute */
  struct days *this;
  int next_start;
  float nominal_charge_save = NOMINAL_CHARGE;
  float minimal_charge_save = MINIMAL_CHARGE;
  float reduction_ratio_save = REDUCTION_RATIO;
  float reduction_second_save = REDUCTION_SECOND;
  int reduction_unit_save = REDUCTION_UNIT;
  
  if(opt_zone_number != 0) {
    error("Option '-help-charge' excludes option '-zone'!\n");
    exit(1);
  }
  printf("Here the charging information for %s\n\
---------------------------------------------\n", COUNTRY);

# if (DAYS_A_WEEK == 1)
  printf("+ Same charging for every weekday\n");
  info_string[0] = "Every weekday";
# elif (DAYS_A_WEEK == 2)
  printf("+ Different charging for weekends (sundays, saturdays)\n  and working days\n");
  info_string[0] = "Weekends";
  info_string[1] = "Working days";
# elif (DAYS_A_WEEK == 3)
  printf("+ Different charging for 1) sundays 2) working days and 3) saturdays\n");
  info_string[0] = "Sundays";
  info_string[1] = "Working days";
  info_string[2] = "Saturdays";
# endif

# ifdef BILLING_PER_UNIT
  printf("+ Charging is based on 'billing per unit' (fixed price\n  \
for time intervals of variable length).\n+ Price of one unit is %s %.3f\n",
	 CURRENCY_NAME, UNIT_PRICE);
  unit_length="unit[sec]";
# else
  printf("+ Charging is based on 'billing per second' (in opposite\n  \
to 'billing per unit').\n");
  unit_length="charge/min";
# endif

  printf("+ The name of the currency is '%s'\n\
+ Charges are printed with %i decimals\n", CURRENCY_NAME, DECIMALS);
  printf("+ tcharge knows this holidays (yeardays in the range 0 to 365):\n  ");
  for(d=1; d <= holidays[0]; d++)
    printf("%i, ", holidays[d]);
  putchar('\n');

  for(z=0; z < NUM_ZONES; z++) {
    /* undo changes eventually made by init_zone[]() */
    NOMINAL_CHARGE = nominal_charge_save;
    MINIMAL_CHARGE = minimal_charge_save;
    REDUCTION_RATIO = reduction_ratio_save;
    REDUCTION_SECOND = reduction_second_save;
    REDUCTION_UNIT = reduction_unit_save;
    init_zone[z](); /* set zone-specific var.s (see charge.h) */

    printf("\n| Zone %i:\n|========\n", z+1);
    for(d=0; d < DAYS_A_WEEK; d++)
      printf("| %-24s", info_string[d]);
    putchar('\n');
    for(d=0; d < DAYS_A_WEEK; d++)
      printf("|-------------------------");
    putchar('\n');
    for(d=0; d < DAYS_A_WEEK; d++)
      printf("| Band         %-11s", unit_length);
    putchar('\n');
    for(b=0; b < TIME_ZONES; b++) {
      for(d=0; d < DAYS_A_WEEK; d++) {
	this = &zones[z][d];
	if(b < this->n) {
	  next_start = (b == (this->n -1)) ? 86400 : this->start[b+1];
      	  printf("| %02i:%02i-%02i:%02i  %-11.3f",
		 this->start[b]/3600, (this->start[b]/60)%60,
		 next_start/3600, (next_start/60)%60, this->length[b]);
	}
	else
	  printf("|                         ");
      }
      putchar('\n');
    }
    for(d=0; d < DAYS_A_WEEK; d++)
      printf("|-------------------------");
    putchar('\n');
    if(NOMINAL_CHARGE)
      printf("+ Each call costs an additional price of %s %.3f\n",
	     CURRENCY_NAME, NOMINAL_CHARGE);
    if(MINIMAL_CHARGE)
      printf("+ Each call costs an minimal price of %s %.3f\n",
	     CURRENCY_NAME, MINIMAL_CHARGE);
    if(REDUCTION_RATIO) {
      if(REDUCTION_SECOND >= 0)
	printf("+ After %.1f seconds ", REDUCTION_SECOND);
      else	/* REDUCTION_UNIT */
	printf("+ After %d units ", REDUCTION_UNIT);
      printf("connection costs are lowered to %.0f%% of the normal price.\n",
	     REDUCTION_RATIO * 100.);
    }
  }
# else
  printf("tcharge: function 'print_charge_structure'\n\
 only available if compiled with -DPRINT_CHARGE_STRUCTURE!\n");
#endif /* PRINT_CHARGE_STRUCTURE */
  return;
}

/*****************************************************************************
 *
 * create a pid file (tcharge_pid_file) if possible
 *  (and `tcharge_pid_file' is not NULL)
 *
 *****************************************************************************/
void write_pid_file(void)
{
  FILE *fp=NULL;
  struct stat buf;
  int file_exists=0;
  
  /* if `tcharge_pid_file' is set to NULL do nothing */
  if(tcharge_pid_file == NULL) return;
  if(lstat(tcharge_pid_file, &buf) == 0) {
    fprintf(stderr, "tcharge: pid file %s\n already exists, trying \
to overwrite it...\n", tcharge_pid_file);
    file_exists = 1;
  }
  if((fp = fopen(tcharge_pid_file, "w")) == NULL)
    fprintf(stderr, "tcharge: Unable to write pid file %s. (PID is %i)\n",
	    tcharge_pid_file, getpid());
  else {
    if(file_exists)
      fprintf(stderr, "tcharge: overwriting existing pid file %s.\n",
	      tcharge_pid_file);
    fprintf(fp, "%i", getpid());
    fclose(fp);
  }
  return;
}

/*****************************************************************************
 *
 *  Extract the resource fields that are needed to open the window.
 *  (geometry and colors)
 *
 *****************************************************************************/
void extractResources(char *geom_string, char *fg_color, char *bg_color)
{
  int x, y, width, height;
  int flags;
  Colormap	colormap;

  colormap = DefaultColormap(display,screen);

  flags = XParseGeometry(geom_string,&x,&y,&width,&height);

  if (flags & WidthValue)
    {
      main_width = width + sizehints.base_width;
      sizehints.flags |= USSize;
    }
  if (flags & HeightValue)
    {
      main_height = height + sizehints.base_height;
      sizehints.flags |= USSize;
    }
  if (flags & XValue)
    {
      if (flags & XNegative)
	{
	  x = MyDisplayWidth + x - main_width - 2*BW;
	  sizehints.win_gravity = NorthEastGravity;
	}
      sizehints.x = x;
      sizehints.flags |= USPosition;
    }
  if (flags & YValue)
    {
      if (flags & YNegative)
	{
	  y = DisplayHeight(display,screen) + y - main_height - 2*BW;
	  sizehints.win_gravity = SouthWestGravity;
	  if((flags&XValue)&&(flags&XNegative))
	    sizehints.win_gravity = SouthEastGravity;
	}
      sizehints.y = y;
      sizehints.flags |= USPosition;
    }

  /*  Do the foreground, and background colors.*/
  if (XParseColor(display,colormap,fg_color,&foreground_color) == 0)
    fprintf(stderr,"tcharge: invalid color %s", fg_color);
  else
    XAllocColor(display,colormap,&foreground_color);

  if (XParseColor(display,colormap,bg_color,&background_color) == 0)
    fprintf(stderr,"tcharge: invalid color %s", bg_color);
  else
    XAllocColor(display,colormap,&background_color);

}

/**************************************************************************
 *
 *  Open and map the windows.
 *
 **************************************************************************/
static Atom wm_del_win;
void createWindows(int argc,char **argv)
{
  XTextProperty wname, iname, mname;
  XClassHint class1, class2;
  XWMHints wmhints;

  sizehints.width = main_width;
  sizehints.height = main_height;
  main_win = XCreateSimpleWindow(display,DefaultRootWindow(display),
				 sizehints.x,sizehints.y,
				 sizehints.width,sizehints.height,
				 BW,foreground_color.pixel,
				 background_color.pixel);
  XStringListToTextProperty(&app_name,1,&wname);

  XStringListToTextProperty(&app_name,1,&iname);

  class1.res_name = app_name;
  class1.res_class = CLASS;
  wmhints.input = True;
  wmhints.flags = InputHint;
  XSetWMProperties(display,main_win,&wname,&iname,argv,argc,
		   &sizehints,&wmhints,&class1);

  XSelectInput(display,main_win,MW_EVENTS);
  XMapWindow(display,main_win);

  /* create, but don't map a window for appointment reminders */
# ifdef REMINDERS
  mess_win = XCreateSimpleWindow(display,DefaultRootWindow(display),
				 sizehints.x,sizehints.y,
				 sizehints.width,sizehints.height,
				 BW,foreground_color.pixel,
				 background_color.pixel);
  if (XStringListToTextProperty(&mess_name,1,&mname) == 0)
    error("cannot allocate window name");
  class2.res_name = mess_name;
  class2.res_class = MESS_CLASS;
  wmhints.input = True;
  wmhints.initial_state = NormalState;
  wmhints.flags = InputHint | StateHint;
  sizehints.flags |= USPosition;
  XSetWMProperties(display,mess_win,&mname,&mname,argv,argc,
		   &sizehints,&wmhints,&class2);

  XSelectInput(display,mess_win,MESS_EVENTS);
# endif /* REMINDERS */

  /* want to accept the delete window protocol */
  wm_del_win = XInternAtom(display,"WM_DELETE_WINDOW",False);
  XSetWMProtocols(display,main_win,&wm_del_win,1);
# ifdef REMINDERS
  XSetWMProtocols(display,mess_win,&wm_del_win,1);
# endif

}

/***************************************************************************
 *
 * Loops forever, looking for stuff to do. Sleeps 'clockupdate' seconds
 * if nothing to do.
 * This version SUPPORTS X-Window
 *
 **************************************************************************/
void getEvent(void)
{
  fd_set in_fdset;
  XEvent event;
  int fd_width = 10;
  struct timeval tm;
  Window root;
  int x,y,border_width,depth;

  for (;;) {
    /* take care of all pending X events */
    while(XPending(display)) {
      XNextEvent(display,&event);
      switch(event.type) {
      case ClientMessage:
	/* check for delete window requests */
	if((event.xclient.format == 32)&&
	   (event.xclient.data.l[0] == wm_del_win)) {
	  if(event.xany.window == main_win)
	    exit(0);
#         ifdef REMINDERS
	  else {
	    XUnmapWindow(display,mess_win);
	    Mapped = 0;
	  }
#         endif
	}
	break;
      case Expose:
      case GraphicsExpose:
	/* need to re-draw a window */
	if(revert_to >= 0) { /* set focus on parent window (only on startup) */
	  XSetInputFocus(display, focus, revert_to, CurrentTime);
	  revert_to = -1;
	}
	if (event.xany.window == main_win)
	  DrawWindow(main_win,main_width,main_height,1,0);
#       ifdef REMINDERS
	else
	  DrawWindow(mess_win,mess_w,mess_h,1,0);
#       endif
	break;
      case ConfigureNotify:
	/* window has been re-sized */
	if (event.xany.window == main_win)
	  XGetGeometry(display,main_win,&root,&x,&y,
		       &main_width,&main_height,&border_width,&depth);
#       ifdef REMINDERS
	else
	  XGetGeometry(display,mess_win,&root,&x,&y,
		       &mess_w,&mess_h,&border_width,&depth);
#       endif
	break;
      case ButtonPress:
#       ifdef REMINDERS
	/* button press to dismiss message window */
	if(event.xany.window == mess_win) {
	  XUnmapWindow(display,mess_win);
	  Mapped = 0;
	}
#       endif
	break;
      }
    }
    
    /* Now wait for time out or new X event */
    FD_ZERO(&in_fdset);
    FD_SET(x_fd,&in_fdset);
    tm.tv_sec = clockupdate;
    tm.tv_usec = 0;
    select(fd_width,&in_fdset,NULL,NULL,&tm);
    
    /* calculate 'total_dur', 'i_units' and/or 'charge';
     * when in "demon mode" look for changes of 'watch_file'
     * return value of CalcUpTime() signals need for full redraw (1) or
     * need to set off an little alarm
     */
    switch (CalcUpTime()) {
    case 0:
      DrawWindow(main_win,main_width,main_height,0,0);
      break;
    case 1:  /* full redraw */
      DrawWindow(main_win,main_width,main_height,1,0);
      break;
    case 2:  /* a "little alarm" */
      if(!opt_bell) {
	DrawWindow(main_win,main_width,main_height,1,1);
	XSync(display,False);
	/* sleep a short while */
	tm.tv_sec = 0;
	tm.tv_usec = BLINKTIME;  /* note: 1/1000 000 seconds! */
	select(0,NULL,NULL,NULL,&tm);
	DrawWindow(main_win,main_width,main_height,1,1);
	XSync(display,False);
      }
      else
	XBell(display, VOLUME);  /* VOLUME: the volume for the bell
				  * in range from -100 to 100 inclusive */
      break;
    }
  }
}

/***************************************************************************
 *
 * Loops forever, looking for stuff to do. Sleeps 'clockupdate' seconds
 * if nothing to do.
 * This is the version WITHOUT X-Window-support
 *
 **************************************************************************/
void getEvent_noX(void)
{
  struct timeval tm;

  for (;;) {
    /* sleep a short while */
    tm.tv_sec = clockupdate;
    tm.tv_usec = 0;
    select(0,NULL,NULL,NULL,&tm);
    
    switch (CalcUpTime()) {
    case 0:
      break;
    case 1:  /* full redraw */
      DrawWindow(main_win,main_width,main_height,1,0);
      break;
    case 2:  /* a "little alarm" */
      putchar(7); /* A no-X bell */
      break;
    }
  }
}
/****************************************************************************
 *
 *  Redraw the whole window after an exposure or size change.
 *  Redraw only the hands after a timeout.
 *  Sound an alarm if needed
 *  Revert video if requested ('rev_video')
 *
 ****************************************************************************/
void DrawWindow(Window win, int width, int height, int full, int rev_video)
{
  int i,j;
  int angle1,angle2,angle3;
  struct ti {
     int s_x,s_y,m_x,m_y,h_x,h_y;
  } at, ot;
  static int cursor_x[2]={0,0}; /* text positions (maybe 2 lines) */
  int cursor_y[2];
  int center_x,center_y;
  static struct ti omt = { -1 };

  if(rev_video) {
    GC t;

#   ifdef DEBUG
    if(current_bg_pixel == foreground_color.pixel)
      printf("Revert to normal!\n");
    else
      printf("Reverse video!\n");
#   endif
    /* switch modes */
    t = rgc;
    rgc = gc;
    gc = t;
    if(current_bg_pixel == foreground_color.pixel)
      current_bg_pixel = background_color.pixel;
    else
      current_bg_pixel = foreground_color.pixel;

    XSetWindowBackground(display,main_win,current_bg_pixel);
    full = 1;
  }    

  if (full)
     XClearWindow(display,win);
  
# ifdef REMINDERS
  if(opt_warn) {
    /* for a message window, just re-draw the message */
    if(win == mess_win) {
      XDrawString (display, mess_win, mgc, 20, 20+font->ascent,
		   message, strlen(message));
      return;
    }
    /* see if we have an alarm to set off */
    switch(opt_warn) {
    case WARN_CHARGE:
      if(charge >= warntime) {
	char warn_charge_str[64];
	
#       ifdef CURRENCY_AFTER_CHARGE
	sprintf(warn_charge_str,
		"This call: Phone charge %%.%df %s", DECIMALS, CURRENCY_NAME);
#       else
	sprintf(warn_charge_str,
		"This call: Phone charge %s %%.%df", CURRENCY_NAME, DECIMALS);
#       endif
	sprintf(message, warn_charge_str, charge);
	SetOffAlarm();
	warntime += opt_warntime;
      }
      break;
    case WARN_UNITS:
      if(i_units >= warntime) {
	sprintf(message, "This call: %d Phone units", i_units);
	SetOffAlarm();
	warntime += opt_warntime;
      }
      break;
    case WARN_MIN:
      if(total_dur/60 >= warntime) {
	sprintf(message, "This call: %d minute(s)", total_dur/60);
	SetOffAlarm();
	warntime += opt_warntime;
      }
      break;
    }
  }
# endif /* REMINDERS */

  if(opt_noclock) {
    height = ((height - text_lines*font_height)>>1);
    if(height < 0)
      height = 0;
    cursor_y[0] = height + font->ascent;
    if(text_lines == 2)
      cursor_y[1] =  cursor_y[0] + font_height;
  }
  else {
    height -= text_lines*font_height+12;
    if(height < 0)
      height = 0;
    cursor_y[0] = height+8+font->ascent;
    if(text_lines == 2)
      cursor_y[1] =  cursor_y[0] + font_height;
  }

  /* draw the textfield */
  if (full || i_units != old_units || charge != old_charge) {
    if(!full) {
      /* erase old text: */
      for(i=0; i < text_lines; i++)
	XDrawString (display, win, rgc, cursor_x[i], cursor_y[i],
		     time_str[i], strlen(time_str[i]));
    }
    old_units = i_units;
    old_charge = charge;

    if(text_lines == 1) {
      if(!halted) {
	if(opt_units == 1)
	  sprintf(time_str[0], "Units: %d", i_units); /* print units only */
	else	/* print charge only */
	  sprintf(time_str[0], charge_str, charge);
      }
      else
	strcpy(time_str[0], "Offline");
    }
    else {	/* print both */
      if(!halted) {
	sprintf(time_str[0], "Units: %d", i_units);
	sprintf(time_str[1], charge_str, charge);
      }
      else {
	strcpy(time_str[0], "Offline");
	strcpy(time_str[1], "- - - -");
      }
    }
    for(i=0; i < text_lines; i++) {
      cursor_x[i] = (width - XTextWidth(font,time_str[i],strlen(time_str[i]))) / 2;
      XDrawString (display, win, gc, cursor_x[i], cursor_y[i],
		   time_str[i],strlen(time_str[i]));
    }
  }

  if(opt_noclock)
    return;

  /* ++++++++++++++++++++++ CLOCK ++++++++++++++++++++++ */
  center_x = width>>1;
  center_y = height>>1;

  if (full) {
     /* draw the clock face */
     for(i=0;i<60;i+=5)
     {
	angle1  = sine[i]*width;
	angle2  = -sine[(i+15)%60]*height;
	XDrawLine(display,win,gc,
		  angle1 * 9 / 20000 + center_x,
		  angle2 * 9 / 20000 + center_y,
		  angle1 *10 / 20000 + center_x,
		  angle2 *10 / 20000 + center_y);
     }
  }

  /* calculate the positions of the hands */
  angle1 = ((total_dur / 3600) % 12) * 5;
  at.h_x =   sine[angle1] * width  *60/200000 + center_x;
  at.h_y =   -(sine[(angle1+15)%60])* height *60/200000 + center_y;

  angle2 = (total_dur / 60) % 60;
  at.m_x =   sine[angle2] * width  *85/200000 + center_x;
  at.m_y = -(sine[(angle2+15)%60])* height *85/200000 + center_y;

  if (clockupdate==1) {
     angle3 = total_dur % 60;
     at.s_x =   sine[angle3] * width  *85/200000 + center_x;
     at.s_y = -(sine[(angle3+15)%60])* height *85/200000 + center_y;
  } else {
     at.s_x = at.s_y = 0;
  }
  ot=omt;
  if (!full && memcmp(&ot,&at,sizeof(struct ti))) {
     for(i=-1;i<2;i++)
	for(j=-1;j<2;j++)
	{
	   XDrawLine(display,win,rgc,center_x+i,center_y+j,ot.h_x,ot.h_y);
	   XDrawLine(display,win,rgc,center_x+i,center_y+j,ot.m_x,ot.m_y);
	}
     if (clockupdate==1) {
	XDrawLine(display,win,rgc,center_x,center_y,ot.s_x,ot.s_y);
     }
     XDrawLine(display,win,rgc,0,height+1,width,height+1);
     XDrawLine(display,win,rgc,0,height+2,width,height+2);
  }

  if (full || memcmp(&ot,&at,sizeof(struct ti))) {
     for(i=-1;i<2;i++)
	for(j=-1;j<2;j++)
	{
	   XDrawLine(display,win,gc,center_x+i,center_y+j,at.h_x,at.h_y);
	   XDrawLine(display,win,gc,center_x+i,center_y+j,at.m_x,at.m_y);
	}
     if (clockupdate==1)
       XDrawLine(display,win,gc,center_x,center_y,at.s_x,at.s_y);
     omt=at;

     /* the line between clock and textfield: */
     XDrawLine(display,win,gc,0,height+1,width,height+1);
     XDrawLine(display,win,gc,0,height+2,width,height+2);
  }
}

/**************************************************************************
 *
 * calculate time connected and units/charge spent so far
 * return value=1 if the window has to be redrawn fully afterwards
 *             =2 if a little alarm has to be set off (=0 else)
 *             =0 else
 *
 **************************************************************************/
int CalcUpTime(void)
{
  register int d, i=0;
  int end_sec, next_start;
  struct days *this=NULL; /* struct days declared in 'charge.h' */
  float sec=0;
  float s;		/* connection duration in a time band */
  int reduction_second_rest = REDUCTION_SECOND;
  int rtn_val=0;	/* return value */
# ifdef BILLING_PER_UNIT
  float units=0.0;	/* float 'units' <-> global int 'i_units' */
# else
  int i_charge;		/* temporary, local var. for rounding charge */
# endif
  time_t now_t;
  struct tm *now;

# ifdef DEBUG
  static int bands_done = 0;
  int bands_done_tmp = 0;
  char *wdays[7] = {"Sun","Mon","Tue","Wed","Thu","Fri","Sat"};
# endif

  if(watch_file != NULL) {
    if(connected()) {
      if(halted) { /* start counting now! (We do not have to set the
		    * starttime since this is done in 'connected()') */
	halted = 0;
	clockupdate = 1; /* we don't want to overlook deleting of 'watch_file'! */
	rtn_val = 2;	/* sound a little alarm */
#       ifdef DEBUG
	printf(" \"Demon mode\": '%s' now exists - starting charge computing.\n",
	       watch_file);
#       endif
      }
    }
    else {
      if(! halted) {
	/* stop counting now! ('clockupdate' is small,
	   we don't have to calculate another time) */
	if(charge > 0.0 && (summaryfile || opt_syslog))
	  print_summary();
	halted = 1;
	clockupdate = clockupdate_save;
	i_units = 0; /* do a reset of charge and units */
	charge = 0.0;
	total_dur = 0;
#       ifdef DEBUG
	printf(" \"Demon mode\": '%s' deleted - stop charge computing.\n",
	       watch_file);
#       endif
	return 2;	/* sound a little alarm */
      }
      else
	return 0;
    }
  }

  /* get the time: */
  if(opt_end)
    now_t = opt_end;
  else
    now_t = time(0);
  now = localtime(&now_t);
  /* (struct tm *) 'start' is initialized in main() */

  if((start->tm_yday != now->tm_yday) && (start->tm_wday >= now->tm_wday)) {
    now->tm_wday += 7;  /* limits duration to approx. one week */
    if(now->tm_wday - start->tm_wday > 6)
      error("call duration exceeds one week (don't do that ;-)");
  }
  charge = NOMINAL_CHARGE;  /* an "offset", usually 0.0 */

  /* The DAY LOOP (for calls overlapping two days) */
  for(d = start->tm_wday, total_dur = 0,
	sec = start->tm_hour*3600 + start->tm_min*60 + start->tm_sec - opt_before;
      d <= now->tm_wday; d++, today = tomorrow) {
#   if (DAYS_A_WEEK == 1)
    this = &zones[opt_zone_number][0];	/* all days */
#   elif (DAYS_A_WEEK == 2)
    if((d%7)%6 == 0 || today)
      this = &zones[opt_zone_number][0];	/* weekends: sunday(0) or saturday(6) */
    else
      this = &zones[opt_zone_number][1];	/* working days */
#   elif (DAYS_A_WEEK == 3)
    if(d%7 == 0 || today)
      this = &zones[opt_zone_number][0];	/* sunday(0) */
    else if((d%7)%6  != 0)
      this = &zones[opt_zone_number][1];	/* working days (1-5) */
    else
      this = &zones[opt_zone_number][2];	/* saturday(6) */
#   endif
    end_sec = (d != now->tm_wday) ?
      86400 : now->tm_hour*3600 + now->tm_min*60 + now->tm_sec + opt_after;
    total_dur += end_sec - sec;
    /* The TIMEZONE LOOP (different bands) */
    for(i=0; i < this->n; i++) {
      next_start = (i == (this->n -1)) ? 86400 : this->start[i+1];
      if(sec > next_start)
	continue;
      if(end_sec > next_start)
	s = (float)next_start - sec;
      else /* finished, break at end of timezone loop */
	s = (float)end_sec - sec;
#     ifdef BILLING_PER_UNIT
      if(REDUCTION_RATIO && REDUCTION_SECOND >= 0) {
	if(reduction_second_rest <= s) {
	  /* REDUCTION_SECONDs are over. This piece of code
	   *   will be reached only once. Let's switch to REDUCTION_UNIT!
	   * Note: There is a difference between doing the translation
	   *   REDUCTION_SECOND -> REDUCTION_UNIT now or in the charge file!
	   */
	  float tmp = units + reduction_second_rest / this->length[i];
	  REDUCTION_UNIT = (int) tmp;
	  if(REDUCTION_SECOND && (REDUCTION_UNIT > tmp || REDUCTION_UNIT == 0))
	    REDUCTION_UNIT++; /* REDUCTION_SECOND=0 -> REDUCTION_UNIT=0 ! */
#         ifdef DEBUG
	  printf("(DEBUG)\t* After %.0f seconds = %d units: price reduces now to %.1f%%\n",
		 REDUCTION_SECOND, (int) tmp, REDUCTION_RATIO*100);
	  printf("\t* Replacing `reduction_second' by `reduction_unit'=%d\n",
		 REDUCTION_UNIT);
#         endif /* DEBUG */
	  REDUCTION_SECOND = -1; /* we don't need it anymore */
	}
	else
	  reduction_second_rest -= s;
      }
      units += s / this->length[i];
#     else /* BILLING_PER_SECOND: */
      if(REDUCTION_RATIO) {
	double this_reduction_ratio;  /* range: [REDUCTION_RATIO .. 1] */
	if(reduction_second_rest > 0) {
	  this_reduction_ratio =
	    (reduction_second_rest < s)   ?
	      ( REDUCTION_RATIO + (1 - REDUCTION_RATIO) * (reduction_second_rest/s) )
		: (1.);	/* 1="still no reduction" */
	  printf("this_reduction_ratio=%.2f, restzeit=%ds, s=%.1f\n",
		 this_reduction_ratio, reduction_second_rest, s);
	}
	else
	  this_reduction_ratio = REDUCTION_RATIO;
	reduction_second_rest -= s;
	charge += (s / 60.0) * this_reduction_ratio * this->length[i];
      }
      else
	charge += (s / 60.0) * this->length[i];
#     endif /* BILLING_PER_UNIT || BILLING_PER_SECOND */

      sec = (float) next_start;

#     ifdef DEBUG
      bands_done_tmp++;
      if(bands_done_tmp > bands_done) { /* remember: this loop runs on every update */
        bands_done++;
        printf("(DEBUG) %s, %d:%02d: Entered 'timezone'[%d] (starting at %d:%02d).\n", 
	       wdays[d%7], (int)sec/3600, ((int)sec/60)%60, i,
	       this->start[i]/3600, (this->start[i]/60)%60);
#       ifdef BILLING_PER_UNIT
        printf("\tDuration of one unit in zone[%d] (price: %.3f %s) is %.1f sec.\n",
	       opt_zone_number, UNIT_PRICE, CURRENCY_NAME, this->length[i]);
        printf("\tCurrent value of units is %.1f (start of|in|end of band?).\n", units);
	if(REDUCTION_RATIO && REDUCTION_UNIT <= units)
	  printf("\tPrice reduction is on.\n");
#       else /* BILLING_PER_SECOND: */
        printf("\tPrice of one minute in zone[%d] is %.2f %s.\n",
	       opt_zone_number, this->length[i], CURRENCY_NAME);
        printf("\tCurrent value of charge is %.2f %s (start of|in|end of band?).\n",
	       charge, CURRENCY_NAME);
#       endif /* BILLING_PER_UNIT || BILLING_PER_SECOND */
        if(next_start != 86400)
	  printf("\tNext 'timezone': no. [%d], starting at %d:%02d.\n", 
		 i+1, this->start[i+1]/3600, (this->start[i+1]/60)%60);
        else
	  printf("\tNext 'timezone' is on next day (%s).\n", wdays[(d+1)%7]);
        if(today)
	  printf("\t-- Public holiday (01/01+%ddays)! Using weekend prices --\n",
		 holidays[today]);
        if(REDUCTION_RATIO) {
	  printf("\tPrice reduction %.0f%% of normal price ",
		 REDUCTION_RATIO*100);
	  if(REDUCTION_SECOND >= 0) {
	    printf("after %.1f seconds\n", REDUCTION_SECOND);
	    if(reduction_second_rest + s > 0)
	      printf("\tSeconds left till reduction starts: %.1f\n",
		     reduction_second_rest + s);
	    else
	      printf("\tPrice reduction on for %.1f seconds.\n",
		     -(reduction_second_rest + s));
	  }
	  else
	    printf("after %d periods\n", REDUCTION_UNIT);
        }
      }
#     endif /* DEBUG */
      if(end_sec <= next_start)
	break;
    } /* END: TIMEZONE LOOP */
    sec = 0;
  } /* END: DAY LOOP */
# ifdef BILLING_PER_UNIT
  i_units = (int)units;
  
  /*
   * See if we have to sound a little alarm before next unit begins
   * -> remaining_seconds = (1.0 - (units-i_units)) * this->length[i]
   *  (i has not been incremented on exit of the 'for(i=...)' loop)
   */
  if(opt_warn_seconds) {
    int tmp = (1.0 - (units-i_units)) * this->length[i] - opt_warn_seconds;
    if(tmp > 0 && tmp <= clockupdate) {
#     ifdef DEBUG
      printf("Option '-warn-seconds %d'. Remaining seconds till next unit starts: %ds\n \
Warn now! (length of a unit: %.1f, clockupdate: %d)\n",
	     opt_warn_seconds, tmp+opt_warn_seconds, this->length[i], clockupdate);
#     endif
      rtn_val = 2;
    }
  }
  if(i_units < units || i_units == 0)
    i_units++;
  if(REDUCTION_UNIT >= 0 && REDUCTION_UNIT < i_units)
    charge =  UNIT_PRICE * (REDUCTION_UNIT  /* normal-price, rest is cheaper: */
			    + (i_units - REDUCTION_UNIT) * REDUCTION_RATIO );
  else
    charge = UNIT_PRICE * i_units;
# else /* BILLING_PER_SECOND */
  /* upward rounding according to DECIMALS: */
  charge *= ten_to_the_decimals;
  if((i_charge = (int)charge) < charge || i_charge == 0)
    i_charge++;
  charge = (float) i_charge / ten_to_the_decimals;
# endif /* BILLING_PER_UNIT || BILLING_PER_SECOND */
  if(charge < MINIMAL_CHARGE)
    charge = MINIMAL_CHARGE;
  
  return rtn_val;
}

/***************************************************************************
 *
 * Check if connection is up
 * set starttime when connected and charge computing is halted (start now)
 *
 **************************************************************************/
int connected(void)
{
# ifndef ISDN
  struct stat statbuf;	/* for status of the file (option -watch-file) */

  if(lstat(watch_file, &statbuf) == 0) {
    if(halted)
      setstarttime(statbuf.st_mtime);
    return 1;
  }
# else /* #ifdef ISDN */
  char buf[8192], *ptr;
  int fd, len;
  /*
   * This code is derived from 'isdnbutton' (cited after 'pppcosts.c').
   */
  /* Check whether any of the isdn network devices is currently active */
  if ((fd = open("/dev/isdninfo",O_RDONLY|O_NDELAY)) >= 0) {
    if ((len = read(fd,buf,sizeof(buf)-1)) >= 0) {
      buf[len] = '\000';
      /* step through 'buf' line for line: */
      printf("buffer is:\n%s\n", buf);
      for (ptr = buf; ptr && *ptr; (ptr = strchr(ptr,'\n')) ? ptr++ : 0) {
	if (!strncmp(ptr,"usage:\t",7))
	  /* if beginning of line matches 'usage:\t': */
	  for (ptr += 7; *ptr && *ptr != '\n'; ptr++) {
	    if (*ptr != '0' && *ptr != ' ') {
	      /** really connected! **/
	      close(fd);
	      if(halted)
		setstarttime (time(0));
	      return 1;
	    }
	  }
      }
    }
    close(fd);
  }
  /* End of stolen code */
# endif /* ISDN */

  return 0;
}

/***************************************************************************
 *
 * Set the start time of a connection
 *
 **************************************************************************/
void setstarttime(time_t start_t)
{
  int i;

  if(opt_start)
    start_t = opt_start;

  /* Scan the time of the tcharge; 'start' is a pointer on 'start_save' */
  start = localtime(&start_t);
  start_save.tm_sec  = start->tm_sec;  start_save.tm_min  = start->tm_min;
  start_save.tm_hour = start->tm_hour; start_save.tm_wday = start->tm_wday;
  start_save.tm_yday = start->tm_yday;
  start = &start_save;

  /* find out (once and for all) if today or tomorrow is a holiday
   * (let's hope a call does not last longer ;-) */
  for(i=1, today=tomorrow=0; i <= holidays[0]; i++) {
    if(start->tm_yday == holidays[i]) {
      today = i; /* for later reference; (i>0) */
      if(++i < holidays[0]) {
	if(start->tm_yday == holidays[i])
	  tomorrow = i;
      }
      break;
    }
  }
  return;
}

/**************************************************************************
 *
 * Sounds an alarm by mapping the message window
 *
 **************************************************************************/
#ifdef REMINDERS
void SetOffAlarm(void)
{
  if(Mapped)
    return;

  XBell(display, VOLUME);  /* VOLUME: the volume for the bell
			    * in range from -100 to 100 inclusive */

  /* compute the window size */
  mess_w=XTextWidth(font,message,strlen(message));
  mess_w += 40;

  mess_h = font_height + 40;

  /* resize and center the window */
  XMoveResizeWindow(display, mess_win, (MyDisplayWidth - mess_w)>>1,
		    (MyDisplayHeight - mess_h)>>1, mess_w, mess_h);

  /* map the sucker */
  XMapRaised(display,mess_win);

  /* Make it noisy */
  XBell(display, VOLUME);

  Mapped = 1;
}
#endif /* REMINDERS */

/**************************************************************************
 *
 *  print_summary() is invoked on termination of a connection
 *  It appends a summary/statistic of the modem connection to the specified
 *  file (if requested on the command line)
 *
 **************************************************************************/
void print_summary (void)
{
  FILE *fp;
  char *wdays[7] = {"Sun","Mon","Tue","Wed","Thu","Fri","Sat"};
  time_t now_t;
  struct tm *now;
  char fprintf_str[128];
# ifdef BILLING_PER_UNIT
#  ifdef CURRENCY_AFTER_CHARGE
  sprintf(fprintf_str, "units: %%d, charge: %%.%df %s\n", DECIMALS, CURRENCY_NAME);
#  else
  sprintf(fprintf_str, "units: %%d, charge: %s %%.%df\n", CURRENCY_NAME, DECIMALS);
#  endif /* CURRENCY_AFTER_CHARGE */
# else /* BILLING_PER_SECOND: */
#  ifdef CURRENCY_AFTER_CHARGE
  sprintf(fprintf_str, "charge: %%.%df %s\n", DECIMALS, CURRENCY_NAME);
#  else
  sprintf(fprintf_str, "charge: %s %%.%df\n", CURRENCY_NAME, DECIMALS);
#  endif /* CURRENCY_AFTER_CHARGE */
# endif /* BILLING_PER_UNIT || BILLING_PER_SECOND */
  
  if(opt_end)
    now_t = opt_end;
  else
    now_t = time(0);
  now = localtime(&now_t);

  if(summaryfile) {
    if((fp = fopen(summaryfile, "a")) == NULL) {
      fprintf(stderr, "tcharge: cannot open '%s' for writing\n", summaryfile);
      return;
    }
#   ifdef DEBUG
    printf("Writing summary of last call to file '%s'\n", summaryfile);
#   endif
    fprintf(fp, "%02d/%02d/%02d %sConnection [%s, %02d:%02d:%02d] \
-> [%s, %02d:%02d:%02d]\n%02d/%02d/%02d   \
duration: %02d:%02d:%02d, ",
	    now->tm_mon+1, now->tm_mday, now->tm_year+1900, user_string,
	    wdays[start->tm_wday], start->tm_hour, start->tm_min, start->tm_sec,
	    wdays[now->tm_wday], now->tm_hour, now->tm_min, now->tm_sec,
	    now->tm_mon+1, now->tm_mday, now->tm_year+1900,
	    total_dur / 3600, (total_dur / 60) % 60, total_dur % 60);
#   ifdef BILLING_PER_UNIT
    fprintf(fp, fprintf_str, i_units, charge);
#   else  /* BILLING_PER_SECOND: */
    fprintf(fp, fprintf_str, charge);
#   endif /* BILLING_PER_UNIT || BILLING_PER_SECOND */
    fclose(fp);
  }
  if (opt_syslog) {     /* wenn usesyslog */
#   ifdef DEBUG
    printf("Send call summary to `syslogd'...\n");
#   endif
    openlog("tcharge", LOG_PID | LOG_NDELAY, LOG_USER);
    syslog(LOG_INFO,
    "%sConnection [%s, %02d:%02d:%02d] -> [%s, %02d:%02d:%02d],\
 duration: %02d:%02d:%02d, ", user_string,
	    wdays[start->tm_wday], start->tm_hour, start->tm_min, start->tm_sec,
	    wdays[now->tm_wday], now->tm_hour, now->tm_min, now->tm_sec,
	    total_dur / 3600, (total_dur / 60) % 60, total_dur % 60);
#   ifdef BILLING_PER_UNIT
    syslog(LOG_INFO, fprintf_str, i_units, charge);
#   else  /* BILLING_PER_SECOND: */
    syslog(LOG_INFO, fprintf_str, charge);
#   endif /* BILLING_PER_UNIT || BILLING_PER_SECOND */
    closelog();
  }

  return;
}

/**************************************************************************
 *
 *  this signal handler
 *  - appends a summary/statistic of the modem connection to the specified
 *    file (if requested on the command line)
 *
 **************************************************************************/
static void signal_handler (int sig)
{
# ifdef DEBUG
  printf("Signal %d caught !\n", sig);
# endif

  if(charge > 0.0 && (summaryfile || opt_syslog))
    print_summary();
    
  signal (sig, SIG_DFL);    /* setting signal handling to default */
  if(tcharge_pid_file != NULL)
    unlink(tcharge_pid_file);
  exit(0);	/* define the exit code (assume all caught signals
		 * should terminate tcharge */
}

/***************************************************************************
 *
 * Error message routine
 *
 **************************************************************************/
void error(char *message)
{
  fprintf(stderr,"tcharge: %s\n",message);
  exit(1);
}
